import { expect } from 'chai';
import sinon from 'sinon';
import { varsnap, Consumer, Producer, Util } from '../varsnap/core.js';
import { consumerConfig, setConsumerConfig, producerConfig, setProducerConfig } from './config_tools.js';
function example(x, y) {
    return x * y;
}
describe('Util', function () {
    describe('getConfig', function () {
        it('can read from env vars', function () {
            const config = Util.getConfig(producerConfig(), 'env');
            expect(config).to.equal('production');
        });
        it('will lowercase configs', function () {
            const configs = {
                'varsnap': true,
                'env': 'TEST',
                'branch': 'test',
            };
            const config = Util.getConfig(configs, 'env');
            expect(config).to.equal('test');
        });
        it('will work with non-string configs', function () {
            const configs = {
                'varsnap': true,
                'env': 'test',
                'branch': 'test',
            };
            const config = Util.getConfig(configs, 'varsnap');
            expect(config).to.equal('true');
        });
        it('will default to empty string if there are no configs', function () {
            const config = Util.getConfig(undefined, 'env'); // eslint-disable-line @typescript-eslint/no-explicit-any
            expect(config).to.equal('');
        });
        it('will default to empty string if config is not defined', function () {
            const configs = {
                'varsnap': true,
                'env': 'test',
                'branch': 'test',
            };
            const config = Util.getConfig(configs, 'producerToken');
            expect(config).to.equal('');
        });
    });
    describe('setConfig', function () {
        it('will convert strings to booleans', function () {
            const config = {};
            Util.setConfig(config, 'varsnap', 'true');
            expect(config.varsnap).to.be.true;
        });
    });
    describe('getSignature', function () {
        it('returns a valid signature', function () {
            const signature = Util.getSignature(Util.getConfig, '');
            expect(signature).to.include('javascript');
            expect(signature).to.include(varsnap.version);
            expect(signature).to.include('getConfig');
        });
        it('returns a signature without implementation logic', function () {
            function example2() { return 123456; }
            expect(example2.toString()).to.include('123456');
            const signature = Util.getSignature(example2, '');
            expect(signature).to.not.include('123456');
        });
        it('returns a signature with trimmed whitespace', function () {
            function example2() { return 5; }
            const signature = Util.getSignature(example2, '');
            expect(signature.slice(0, 1)).to.not.equal(' ');
            expect(signature.slice(-1)).to.not.equal(' ');
        });
        it('strips whitespace from signature', function () {
            const signature = Util.getSignature(example, '');
            expect(signature).to.include('example');
            expect(signature).to.include('x,y');
        });
        it('allows a custom funcName', function () {
            const signature = Util.getSignature(Util.getConfig, 'asdf');
            expect(signature).to.include('asdf');
            expect(signature).to.not.include('getConfig');
        });
    });
    describe('debug', function () {
        it('does not debug if no logger is provided', function () {
            const config = {};
            Util.debug(config, 'debug');
        });
    });
    describe('log', function () {
        it('does not log if no logger is provided', function () {
            const config = {};
            Util.log(config, 'log');
        });
    });
    describe('error', function () {
        it('does not error if no logger is provided', function () {
            const config = {};
            Util.error(config, 'error');
        });
    });
    describe('getProducers', function () {
        it('returns a list of producers', function () {
            let producers = Util.getProducers();
            expect(producers).to.have.lengthOf(0);
            varsnap(example);
            producers = Util.getProducers();
            expect(producers).to.have.lengthOf(1);
        });
    });
    describe('limitString', function () {
        it('returns a short string', function () {
            const x = 'a';
            expect(Util.limitString(x)).to.equal(x);
            const y = 'a'.repeat(30);
            expect(Util.limitString(y)).to.equal(y);
        });
        it('returns a shortened string', function () {
            const x = 'a'.repeat(40);
            const limited = Util.limitString(x);
            expect(limited.length).to.equal(30);
            expect(limited.substring(0, 27)).to.equal('a'.repeat(27));
            expect(limited.substring(27, 30)).to.equal('...');
        });
    });
    describe('isNode', function () {
        it('returns the correct value', function () {
            let isBrowser = true;
            if (typeof navigator === 'undefined') {
                isBrowser = false;
            }
            else if (navigator.userAgent.includes('Node.js')) {
                isBrowser = false;
            }
            expect(Util.isNode()).to.not.equal(isBrowser);
        });
    });
});
describe('Producer', function () {
    beforeEach(function () {
        this.producer = new Producer(example, 'example', producerConfig());
    });
    describe('isEnabled', function () {
        it('will return false if varsnap is not enabled', function () {
            this.producer.config.varsnap = false;
            expect(this.producer.isEnabled()).to.be.false;
        });
        it('will return false if env is not production', function () {
            this.producer.config.env = 'development';
            expect(this.producer.isEnabled()).to.be.false;
        });
        it('will return false if producer token is not set', function () {
            this.producer.config.producerToken = '';
            expect(this.producer.isEnabled()).to.be.false;
        });
        it('will return true if all configs are correct', function () {
            expect(this.producer.isEnabled()).to.be.true;
        });
    });
    describe('serialize', function () {
        it('can serialize data', function () {
            expect(Producer.serialize(5)).to.equal('5');
            expect(Producer.serialize({ '1': [2] })).to.equal('{"1":[2]}');
        });
    });
    describe('serialize_formatted', function () {
        it('can serialize data with formatting', function () {
            expect(Producer.serialize_formatted(5)).to.equal('5');
            expect(Producer.serialize_formatted({ '1': [2] })).to.equal('{\n  "1": [\n    2\n  ]\n}');
        });
    });
    describe('produce', function () {
        beforeEach(function () {
            setProducerConfig();
            this.ajax = sinon.stub(Util, 'ajax');
        });
        afterEach(function () {
            this.ajax.restore();
        });
        it('will send data', function () {
            this.ajax.resolves('asdf');
            const producerPromise = this.producer.produce([2, 3], 6);
            return producerPromise.then((produceStatus) => {
                expect(produceStatus).to.be.true;
                sinon.assert.called(this.ajax);
            });
        });
        it('will deal with 429 errors', function () {
            const axiosError = { response: { status: 429 } };
            this.ajax.rejects(axiosError);
            const producerPromise = this.producer.produce([2, 3], 6);
            return producerPromise.then((produceStatus) => {
                expect(produceStatus).to.be.true;
                sinon.assert.called(this.ajax);
            });
        });
        it('will deal with 404 errors', function () {
            this.ajax.rejects({ statusCode: 404 });
            const producerPromise = this.producer.produce([2, 3], 6);
            return producerPromise.then((produceStatus) => {
                expect(produceStatus).to.be.false;
                sinon.assert.called(this.ajax);
            });
        });
        it('will deal with other errors', function () {
            this.ajax.rejects({ statusCode: 500 });
            const producerPromise = this.producer.produce([2, 3], 6);
            return producerPromise.then((produceStatus) => {
                expect(produceStatus).to.be.false;
                sinon.assert.called(this.ajax);
            });
        });
    });
});
describe('Consumer', function () {
    beforeEach(function () {
        this.consumer = new Consumer(example, 'example', consumerConfig());
    });
    describe('isEnabled', function () {
        it('will return false if varsnap is not enabled', function () {
            this.consumer.config.varsnap = false;
            expect(this.consumer.isEnabled()).to.be.false;
        });
        it('will return false if env is not development', function () {
            this.consumer.config.env = 'production';
            expect(this.consumer.isEnabled()).to.be.false;
        });
        it('will return false if consumer token is not set', function () {
            this.consumer.config.consumerToken = '';
            expect(this.consumer.isEnabled()).to.be.false;
        });
        it('will return true if all configs are correct', function () {
            expect(this.consumer.isEnabled()).to.be.true;
        });
    });
    describe('deserialize', function () {
        it('can deserialize data', function () {
            expect(Consumer.deserialize('5')).to.equal(5);
            expect(Consumer.deserialize('{"1":[2]}')).to.deep.equal({ '1': [2] });
        });
    });
    describe('consume', function () {
        beforeEach(function () {
            setConsumerConfig();
            this.ajax = sinon.stub(Util, 'ajax');
            this.trialGroup = {
                status: 'ok',
                project_id: 'project_id',
                trial_group_id: 'trial_group_id',
                trial_group_url: 'trial_group_url',
            };
        });
        afterEach(function () {
            this.ajax.restore();
        });
        it('will not run if the consumer is not enabled', function () {
            this.consumer.config.env = 'production';
            this.consumer.consume(this.trialGroup);
            sinon.assert.callCount(this.ajax, 0);
        });
        it('will make an ajax request and report it', function () {
            const data = {
                'id': 'abcd',
                'inputs': '[2, 3]',
                'prod_outputs': '4',
            };
            this.ajax.resolves(data);
            this.consumer.consume(this.trialGroup);
            sinon.assert.called(this.ajax);
        });
    });
});
describe('varsnap', function () {
    beforeEach(function () {
        setProducerConfig();
        Util.resetConsumers();
        Util.resetProducers();
    });
    it('creates a consumer and producer', function () {
        const wrapped = varsnap(example);
        expect(wrapped.producer.targetFunc).to.equal(example);
        expect(wrapped.producer.config).to.equal(varsnap._config);
        expect(wrapped.consumer.targetFunc).to.equal(example);
        expect(wrapped.consumer.config).to.equal(varsnap._config);
    });
    it('wraps a function', function () {
        const wrapped = varsnap(example);
        expect(wrapped(2, 4)).to.equal(8);
    });
    it('saves consumers and producers', function () {
        expect(Util.getConsumers().length).to.equal(0);
        expect(Util.getConsumers().length).to.equal(0);
        varsnap(example);
        expect(Util.getConsumers().length).to.equal(1);
        expect(Util.getConsumers().length).to.equal(1);
    });
});
describe('version', function () {
    it('has a version', function () {
        expect(varsnap.version).to.be;
    });
});
//# sourceMappingURL=core.js.map