import { expect } from 'chai';
import varsnap from '../varsnap/index.js';
import { Util } from '../varsnap/core.js';
import { FakeLogger } from './config_tools.js';
function generateReferenceFunction() {
    const f = function example1(x, y) {
        return x * y;
    };
    return f;
}
function generateTestFunction() {
    const f = function example1(x, y) {
        return x * y + 1;
    };
    return f;
}
function generateExceptionFunction() {
    const f = function example2() {
        throw 'error';
    };
    return f;
}
function generateUnknownFunction() {
    const f = function example3(x, y) {
        return x * y;
    };
    const randomName = Math.ceil(Math.random() * Math.pow(10, 10)).toString();
    f.toString = function () { return randomName; };
    Object.defineProperty(f, 'name', { value: randomName });
    return f;
}
describe('TestTest', function () {
    describe('generateReferenceFunction', function () {
        it('generates a function', function () {
            const f = generateReferenceFunction();
            expect(f(2, 3)).to.equal(6);
            expect(f(4, 5)).to.equal(20);
            expect(f(6, 7)).to.equal(42);
        });
    });
    describe('generateTestFunction', function () {
        it('generates a function', function () {
            const f = generateTestFunction();
            expect(f(2, 3)).to.equal(7);
            expect(f(4, 5)).to.equal(21);
            expect(f(6, 7)).to.equal(43);
        });
        it('does not match output', function () {
            const f1 = generateReferenceFunction();
            const f2 = generateTestFunction();
            expect(f1(2, 3)).to.not.equal(f2(2, 3));
        });
        it('matches signature', function () {
            const f1 = generateReferenceFunction();
            const f2 = generateTestFunction();
            expect(Util.getSignature(f1, '')).to.include('example');
            expect(Util.getSignature(f1, '')).to.equal(Util.getSignature(f2, ''));
        });
    });
    describe('generateExceptionFunction', function () {
        it('throws an exception', function () {
            const f = generateExceptionFunction();
            expect(() => f(2, 3)).to.throw('error');
        });
    });
    describe('generateUnknownFunction', function () {
        it('has a random signature', function () {
            const f1 = generateUnknownFunction();
            const f2 = generateUnknownFunction();
            expect(Util.getSignature(f1, '')).to.not.equal(Util.getSignature(f2, ''));
        });
    });
});
describe('varsnap', function () {
    function setProductionVarsnap() {
        varsnap.updateConfig({
            varsnap: true,
            env: 'production',
            branch: 'test',
            producerToken: 'producer-o351gr57h2ejznwhrhmu',
            logger: FakeLogger,
        });
    }
    function setDevelopmentVarsnap() {
        varsnap.updateConfig({
            varsnap: true,
            env: 'development',
            branch: 'test',
            consumerToken: 'consumer-053p9sitlfmbkv4mpeoa',
            logger: FakeLogger,
        });
    }
    function resetVarsnap() {
        varsnap.updateConfig({
            varsnap: false,
            env: 'development',
            branch: 'test',
            producerToken: 'producer-abcd',
            consumerToken: 'consumer-abcd',
        });
    }
    this.timeout(5 * 1000);
    beforeEach(function () {
        resetVarsnap();
        varsnap.reset();
    });
    afterEach(function () {
        resetVarsnap();
        varsnap.reset();
    });
    it('can wrap and test a function', function () {
        const f = varsnap(generateReferenceFunction());
        setProductionVarsnap();
        f(2, 3);
        f(4, 5);
        f(6, 7);
        setDevelopmentVarsnap();
        return varsnap.runTests().then(status => {
            expect(status).to.be.true;
        });
    });
    it('can wrap and test a broken function', function () {
        const f1 = varsnap(generateReferenceFunction());
        setProductionVarsnap();
        f1(2, 3);
        f1(4, 5);
        f1(6, 7);
        resetVarsnap();
        varsnap.reset();
        varsnap(generateTestFunction());
        setDevelopmentVarsnap();
        return varsnap.runTests().then(status => {
            expect(status).to.be.false;
        });
    });
    it('can no-op in production', function () {
        const f = varsnap(generateReferenceFunction());
        resetVarsnap();
        expect(f(2, 3)).to.equal(6);
    });
    it('can no-op in development', function () {
        varsnap(generateReferenceFunction());
        resetVarsnap();
        return varsnap.runTests().then(status => {
            expect(status).to.be.false;
        });
    });
    it('can test functions that throw exceptions', function () {
        const f = varsnap(generateExceptionFunction());
        setProductionVarsnap();
        expect(() => f(2, 3)).to.throw();
        expect(() => f(4, 5)).to.throw();
        expect(() => f(6, 7)).to.throw();
        setDevelopmentVarsnap();
        return varsnap.runTests().then(status => {
            expect(status).to.be.true;
        });
    });
    it('will warn if there are no available snaps', function () {
        varsnap(generateUnknownFunction());
        setDevelopmentVarsnap();
        return varsnap.runTests().then(status => {
            expect(status).to.be.true;
            // TODO - check logs to make sure a warning message is emitted
        });
    });
});
//# sourceMappingURL=integration.js.map