import axios from 'axios';
const produceURL = 'https://www.varsnap.com/api/snap/produce/';
const consumeURL = 'https://www.varsnap.com/api/snap/consume/';
const produceTrialURL = 'https://www.varsnap.com/api/trial/produce/';
const trialGroupURL = 'https://www.varsnap.com/api/trial_group/';
export const version = '1.11.3';
export const clientName = 'javascript.' + version;
const defaultConfig = {
    varsnap: true,
    env: 'development',
    branch: '',
    producerToken: '',
    consumerToken: '',
    logger: console,
};
let Producers = [];
let Consumers = [];
export class Util {
    static getConfig(config, configKey) {
        if (!config) {
            return '';
        }
        let val = config[configKey];
        if (!val) {
            return '';
        }
        val = val.toString().toLowerCase();
        return val;
    }
    static setConfig(config, configKey, configValue) {
        if (configKey === 'varsnap') {
            if (typeof configValue === 'string') {
                configValue = configValue === 'true';
            }
            config[configKey] = configValue;
        }
        if (configKey === 'env' || configKey === 'producerToken' || configKey === 'consumerToken' || configKey === 'branch') {
            config[configKey] = configValue;
        }
        if (configKey === 'logger') {
            config[configKey] = configValue;
        }
        return config;
    }
    static getSignature(targetFunc, funcName) {
        let signature = '';
        if (funcName !== '' && funcName !== undefined) {
            signature = funcName;
        }
        else {
            // Remove any implementation code from signature
            signature = targetFunc.toString().split('{')[0];
            // Trim preceding and trailing whitespace
            signature = signature.replace(/^\s+|\s+$/g, '');
            // Trim whitespace between comma separated arguments
            signature = signature.replace(/,\s+/g, ',');
        }
        return 'javascript.' + version + '.' + signature;
    }
    static ajax(url, requestData) {
        const requestParams = new URLSearchParams(requestData);
        const data = requestParams.toString();
        const options = {
            method: 'POST',
            url: url,
            headers: {
                'Content-Type': 'application/x-www-form-urlencoded',
            },
            data: data,
            responseType: 'json',
        };
        if (Util.isNode() && options.headers) {
            options.headers['User-Agent'] = 'varsnap.js';
        }
        const ajaxPromise = axios(options).then((response) => {
            return response.data;
        }).catch((error) => {
            return error;
        });
        return ajaxPromise;
    }
    static debug(config, data) {
        if (config.logger) {
            config.logger.debug(data);
        }
    }
    static log(config, data) {
        if (config.logger) {
            config.logger.log(data);
        }
    }
    static error(config, data) {
        if (config.logger) {
            config.logger.error(data);
        }
    }
    static resetProducers() {
        Producers = [];
    }
    static resetConsumers() {
        Consumers = [];
    }
    static getProducers() {
        return Producers;
    }
    static getConsumers() {
        return Consumers;
    }
    static limitString(x) {
        const limitedString = String(x);
        const limit = 30;
        const ellipsis = '...';
        if (limitedString.length <= limit) {
            return limitedString;
        }
        return limitedString.substring(0, 30 - ellipsis.length) + ellipsis;
    }
    static isNode() {
        if (typeof process !== 'object') {
            return false;
        }
        if (typeof window !== 'undefined') {
            return false;
        }
        return true;
    }
}
export class Producer {
    constructor(targetFunc, signature, config) {
        this.targetFunc = targetFunc;
        this.signature = signature;
        this.config = config;
    }
    isEnabled() {
        if (!this.config.varsnap) {
            return false;
        }
        if (Util.getConfig(this.config, 'env') !== 'production') {
            return false;
        }
        if (!Util.getConfig(this.config, 'producerToken')) {
            return false;
        }
        return true;
    }
    static serialize(data) {
        return JSON.stringify(data);
    }
    static serialize_formatted(data) {
        return JSON.stringify(data, null, 2);
    }
    produce(args, output) {
        if (!this.isEnabled()) {
            const errorPromise = new Promise((resolve) => {
                return resolve(false);
            });
            return errorPromise;
        }
        Util.debug(this.config, 'Sending call to Varsnap');
        return Promise.resolve(output).then((outputValue) => {
            const data = {
                'producer_token': Util.getConfig(this.config, 'producerToken'),
                'signature': this.signature,
                'inputs': Producer.serialize(args),
                'prod_outputs': Producer.serialize(outputValue),
            };
            return Util.ajax(produceURL, data).then(() => {
                return true;
            }).catch((err) => {
                if (err.response) {
                    if (err.response.status === 429) {
                        Util.error(this.config, 'Cannot produce to Varsnap; Signature at rate limit');
                        return true;
                    }
                    if (err.response.status === 404) {
                        Util.error(this.config, 'Cannot produce to Varsnap; check your producer token');
                    }
                    else {
                        Util.error(this.config, err.response.statusText);
                    }
                }
                return false;
            });
        });
    }
}
export class Consumer {
    constructor(targetFunc, signature, config) {
        this.targetFunc = targetFunc;
        this.signature = signature;
        this.config = config;
    }
    isEnabled() {
        if (!this.config.varsnap) {
            return false;
        }
        if (Util.getConfig(this.config, 'env') !== 'development') {
            return false;
        }
        if (!Util.getConfig(this.config, 'consumerToken')) {
            return false;
        }
        return true;
    }
    static deserialize(data) {
        if (data === '') {
            return undefined;
        }
        return JSON.parse(data);
    }
    consume(trialGroup) {
        if (!this.isEnabled()) {
            const errorPromise = new Promise((resolve) => {
                Util.error(this.config, 'Varsnap not enabled for testing');
                return resolve(false);
            });
            return errorPromise;
        }
        const data = {
            'consumer_token': Util.getConfig(this.config, 'consumerToken'),
            'signature': this.signature,
        };
        const ajax = Util.ajax(consumeURL, data);
        const consume = ajax.then((rawResponse) => {
            const response = rawResponse;
            if (!response || response.status !== 'ok') {
                const errorPromise = new Promise((resolve) => {
                    Util.error(this.config, 'Received error from Varsnap: ' + response.status);
                    return resolve(false);
                });
                return [errorPromise];
            }
            if (response.results.length === 0) {
                Util.log(this.config, 'No snaps found for signature: ' + data.signature);
            }
            const trialPromises = [];
            for (const snap of response.results) {
                const inputs = Consumer.deserialize(snap['inputs']);
                const prodOutputs = Consumer.deserialize(snap['prod_outputs']);
                let localOutputs;
                try {
                    localOutputs = this.targetFunc(...inputs);
                }
                catch (err) {
                    localOutputs = err;
                }
                const trialPromise = Promise.resolve(localOutputs).then((localOutputValue) => {
                    const prodOutputsSerialized = Producer.serialize(prodOutputs);
                    const localOutputsSerialized = Producer.serialize(localOutputValue);
                    const matches = prodOutputsSerialized === localOutputsSerialized;
                    this.reportCentral(trialGroup, snap, prodOutputs, localOutputValue, matches).then((trialURL) => {
                        this.reportLog(snap['id'], inputs, prodOutputs, localOutputValue, matches, trialURL);
                    });
                    return matches;
                });
                trialPromises.push(trialPromise);
            }
            return trialPromises;
        }).then((trialPromises) => {
            return Promise.all(trialPromises).then((results) => {
                const allMatches = results.every((result) => result);
                return allMatches;
            });
        }).catch((error) => {
            Util.error(this.config, 'Cannot contact Varsnap: ' + error);
            return false;
        });
        return consume;
    }
    reportCentral(trialGroup, snap, prodOutputs, localOutputs, matches) {
        const data = {
            'trial_group_id': trialGroup.trial_group_id,
            'consumer_token': Util.getConfig(this.config, 'consumerToken'),
            'snap_id': snap.id,
            'test_outputs': Producer.serialize(localOutputs),
            'matches': matches.toString(),
            'snap_outputs_formatted': Producer.serialize_formatted(prodOutputs),
            'test_outputs_formatted': Producer.serialize_formatted(localOutputs),
        };
        const ajax = Util.ajax(produceTrialURL, data);
        const report = ajax.then((rawResponse) => {
            const response = rawResponse;
            if (!response || response.status !== 'ok') {
                return false;
            }
            const trialURL = response['trial_url'];
            return trialURL;
        }).catch((error) => {
            Util.error(this.config, 'Cannot upload test results to Varsnap: ' + error);
            return false;
        });
        return report;
    }
    reportLog(snapID, inputs, prodOutputs, localOutputs, matches, trialURL) {
        if (matches) {
            return;
        }
        Util.log(this.config, '');
        Util.log(this.config, 'Testing with Varsnap uuid:     ' + snapID);
        Util.log(this.config, 'Signature:                     ' + this.signature);
        Util.log(this.config, 'Function input args:           ' + Util.limitString(inputs));
        Util.log(this.config, 'Production function outputs:   ' + Util.limitString(prodOutputs));
        Util.log(this.config, 'Your function outputs:         ' + Util.limitString(localOutputs));
        Util.log(this.config, 'Matching outputs:              ' + matches);
        let trialURLView = 'Cannot upload trial';
        if (trialURL) {
            trialURLView = String(trialURL);
        }
        Util.log(this.config, 'Trial URL:                     ' + trialURLView);
    }
    reportTrialGroup(trialGroup) {
        Util.log(this.config, 'Trial Group:                   ' + trialGroup.trial_group_url);
    }
}
export function varsnap(func, funcName) {
    if (funcName === undefined) {
        funcName = '';
    }
    const signature = Util.getSignature(func, funcName);
    const producer = new Producer(func, signature, varsnap._config);
    const consumer = new Consumer(func, signature, varsnap._config);
    const wrapped = function (...args) {
        let result;
        try {
            result = func(...args);
        }
        catch (err) {
            producer.produce(args, err);
            throw (err);
        }
        producer.produce(args, result);
        return result;
    };
    wrapped.producer = producer;
    wrapped.consumer = consumer;
    wrapped.signature = signature;
    Producers.push(producer);
    Consumers.push(consumer);
    return wrapped;
}
varsnap._config = defaultConfig;
varsnap.updateConfig = function updateConfig(newConfig) {
    let config = varsnap._config;
    for (const key of Object.keys(varsnap._config)) {
        if (Object.prototype.hasOwnProperty.call(newConfig, key)) {
            config = Util.setConfig(config, key, newConfig[key]);
        }
    }
    varsnap._config = config;
};
varsnap.reset = function reset() {
    Util.resetConsumers();
    Util.resetProducers();
};
varsnap.version = version;
function getTrialGroup(consumerConfig) {
    const data = {
        'branch': Util.getConfig(consumerConfig, 'branch'),
        'consumer_token': Util.getConfig(consumerConfig, 'consumerToken'),
        'client': clientName,
    };
    const raw = data;
    const ajax = Util.ajax(trialGroupURL, raw);
    const trialGroup = ajax.then((rawResponse) => {
        const response = rawResponse;
        return response;
    }).catch((error) => {
        let errorMessage = '';
        if (error && error.response && error.response.data) {
            const response = error.response.data;
            errorMessage = response.error;
        }
        else {
            errorMessage = error.toString();
        }
        Util.error(consumerConfig, 'Cannot generate a new Trial Group: ' + errorMessage);
        return {
            status: 'not ok',
            project_id: '',
            trial_group_id: '',
            trial_group_url: '',
        };
    });
    return trialGroup;
}
export function runTests() {
    const consumers = Util.getConsumers();
    if (consumers.length === 0) {
        return Promise.resolve(true);
    }
    const consumerConfig = consumers[0].config;
    return getTrialGroup(consumerConfig).then((trialGroup) => {
        const testPromises = [];
        for (const consumer of consumers) {
            const testPromise = consumer.consume(trialGroup);
            testPromises.push(testPromise);
        }
        if (consumers.length > 0) {
            consumers[0].reportTrialGroup(trialGroup);
        }
        return Promise.all(testPromises);
    }).then((results) => {
        const allPass = results.every((result) => result);
        return allPass;
    });
}
varsnap.runTests = runTests;
//# sourceMappingURL=core.js.map